;;; -*- Mode: LISP; Syntax: COMMON-LISP; Package: ALPINE; Base: 10 -*-
;;;
;;;  (c) copyright 2007-2008 by
;;;           Samium Gromoff (_deepfire@feelingofgreen.ru)
;;;
;;; This library is free software; you can redistribute it and/or
;;; modify it under the terms of the GNU Library General Public
;;; License as published by the Free Software Foundation; either
;;; version 2 of the License, or (at your option) any later version.
;;;
;;; This library is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;;; Library General Public License for more details.
;;;
;;; You should have received a copy of the GNU Library General Public
;;; License along with this library; if not, write to the
;;; Free Software Foundation, Inc., 59 Temple Place - Suite 330,
;;; Boston, MA  02111-1307  USA.

(in-package :alpine)

(defclass mount ()
  ((name :accessor name :type string :initarg :name)
   (height :accessor height :type integer :initarg :height)
   (location :accessor location :initarg :location)
   (routes :reader routes :initform (make-hash-table :test 'equal))))

(eval-when (:compile-toplevel :load-toplevel :execute)
  (defvar *route-grades* '(1b 2a 2b 3a 3b 4a 4b 5a 5b 6a 6b 7a 7b)))

(defun route-grade> (a b)
  (> (position a *route-grades*) (position b *route-grades*)))

(deftype route-grade () `(member ,*route-grades*))

(defclass route ()
  ((mount :accessor route-mount :type mount :initarg :mount)
   (name :accessor name :type string :initarg :name)
   (grade :accessor route-grade :type route-grade :initarg :grade)))

(defclass ascent ()
  ((route :accessor ascent-route :type route :initarg :route)
   (date :accessor date :type integer :initarg :date)
   (members :accessor members :type list :initarg :members)))

(deftype name () 'string)
(deftype surname () 'string)
(deftype familyname () 'string)
(deftype date () '(integer 0))
(deftype address () 'string)
(deftype phone-number () '(integer 0))
(deftype aim-id () 'string)
(deftype email () 'string)
(deftype passport-id () '(integer 0))
(deftype passport-issuer () 'string)
(deftype form () '(or cons symbol number))

(define-presentation-type url () :inherit-from 'string)
(define-presentation-type name () :inherit-from 'string)
(define-presentation-type surname () :inherit-from 'string)
(define-presentation-type familyname () :inherit-from 'string)
(define-presentation-type date () :inherit-from 'integer)
(define-presentation-type address () :inherit-from 'string)
(define-presentation-type phone-number () :inherit-from 'integer)
(define-presentation-type aim-id () :inherit-from 'string)
(define-presentation-type email () :inherit-from 'string)
(define-presentation-type passport-id () :inherit-from 'integer)
(define-presentation-type passport-issuer () :inherit-from 'string)
(define-presentation-type alpinist-grade () :inherit-from 'string)
(define-presentation-type query () :inherit-from 'string)

(defclass human ()
  ((id :accessor id :type symbol :initarg :id :documentation "Id")
   (flag :accessor flag :type boolean :initform nil :initarg :flag)))

(deftype alpinist-grade () '(member novice ensign 3 2 1 kms ms msmk zms))

(defclass alpinist (human)
  ((ascents :accessor ascents :type list :initform nil :initarg :ascents)))

(defclass flag ()
  ((who :accessor who :initarg :who)))

(defclass standard-alpinist (alpinist) ())

(defclass event ()
  ((controller :accessor controller :type standard-alpinist :initarg :controller)
   (ascents :accessor ascents :type list :initarg :ascents)))

(defclass key ()
  ((id :accessor id :type symbol :initarg :id)
   (description :accessor description :type string :initarg :description)
   (data :accessor key-data :initform (make-hash-table :test 'equal) :type hash-table)
   (value-type :accessor key-value-type :initform nil :initarg :value-type)))

(defclass standard-key (key) ())
(defclass standard-key-banner (standard-key) ())
(defclass metakey (key) ())
(defclass pure-metakey (metakey) ())

(defmethod print-object ((o key) stream)
  (print-unreadable-object (o stream)
    (format stream "~S ~S" (id o) (description o))))

(defclass key-value ()
  ((key :accessor value-key :initarg :key)
   (key-key :accessor value-key-key :initarg :key-key)))

(defclass unset-key-value (key-value) ())

(defclass brief-key-value (key-value) ())
(defclass brief-unset-key-value (key-value) ())

(defclass realm ()
  ((mounts :accessor realm-mounts :type hash-table :initform (make-hash-table :test 'equal))
   (routes :accessor realm-routes :type hash-table :initform (make-hash-table :test 'equal))
   (alpinists :accessor realm-alpinists :type hash-table :initform (make-hash-table :test 'equal))
   (names :accessor realm-name# :type hash-table :initform (make-hash-table :test 'equal))
   (storage-pathname :accessor realm-storage-pathname :initform nil :type pathname)))

(define-condition undefined-realm-object (error)
  ((realm :accessor condition-realm :initarg :realm)
   (id :accessor condition-id :initarg :id))
  (:report (lambda (c s)
             (format s "~@<Realm ~S has no object named ~S.~:@>" (condition-realm c) (condition-id c)))))

(define-condition conflicting-realm-object (error)
  ((realm :accessor condition-realm :initarg :realm)
   (id :accessor condition-id :initarg :id))
  (:report (lambda (c s)
             (format s "~@<Realm ~S already has an object named ~S.~:@>" (condition-realm c) (condition-id c)))))

(defun realm-object (realm id &key (error-p t))
  (or (gethash id (realm-name# realm))
      (when error-p
        (error 'undefined-realm-object :realm realm :id id))))

(defun (setf realm-object) (object realm id &key error-p)
  (if (and error-p (gethash id (realm-name# realm)))
      (error 'conflicting-realm-object :realm realm :id id)
      (setf (gethash id (realm-name# realm)) object)))

(defun map-realm-objects (fn realm)
  (maphash-values fn (realm-name# realm)))

(defun remove-realm-object (realm id)
  (remhash id (realm-name# realm)))

(defun realm-add-key (realm class id description)
  (setf (realm-object realm id :error-p t) (make-instance class :id id :description description)))

(defun realm-key (realm id)
  (realm-object realm id))

(defun map-realm-keys (fn type realm)
  (maphash-values fn (realm-name# realm) :test (of-type type)))

(defun the-key-value (key id)
  (declare (type key key) (type symbol id))
  (gethash id (key-data key)))

(defun (setf the-key-value) (val key id)
  (declare (type key key) (type symbol id))
  (setf (gethash id (key-data key)) val))

(defun (setf the-key-values) (val key ids)
  (mapcar (curry #'(setf the-key-value) val key) ids))

(defun (setf the-key-values-*) (val key &rest ids)
  (setf (the-key-values key ids) val))

(defun map-the-key (fn key &key (test #'identity))
  (maphash* fn (key-data key) :test test))

(defun map-metakey (fn key)
  (maphash-keys fn (key-data key) :test #'identity))

(defun realm-key-value (realm key-id id)
  (declare (type symbol key-id id))
  (the-key-value (realm-key realm key-id) id))

(defun (setf realm-key-value) (val realm key-id id)
  (declare (type symbol key-id id))
  (setf (the-key-value (realm-key realm key-id) id) val))

(defmethod initialize-instance :after ((o realm) &key &allow-other-keys)
  (realm-add-key o 'metakey 'displayed-keys "Displayed keys"))

(defvar *realm* (make-instance 'realm))

(define-container-hash-accessor *realm* mount :container-transform realm-mounts)
(define-container-hash-accessor *realm* route :container-transform realm-routes)
(define-container-hash-accessor *realm* alpinist :container-transform realm-alpinists :mapper t)

(defun alpinist-key-value (a key)
  (realm-key-value *realm* key (id a)))

(defun (setf alpinist-key-value) (val a key)
  (setf (realm-key-value *realm* key (id a)) val))

(defun alpinist-key-value* (value)
  (realm-key-value *realm* (id (value-key value)) (value-key-key value)))

(defun (setf alpinist-key-value*) (val value)
  (setf (realm-key-value *realm* (id (value-key value)) (value-key-key value)) val))

(defun alpinist-member-keys (id)
  (remove-if-not (curry #'gethash id) (map-realm-keys #'identity 'standard-key *realm*) :key #'key-data))

(defun alpinist-non-member-keys (id)
  (remove-if (curry #'gethash id) (map-realm-keys #'identity 'standard-key *realm*) :key #'key-data))

(defun metakey- (keys &rest metakeys)
  (labels ((do- (acc rest)
             (if (null rest)
                 acc
                 (do- (set-difference keys (map-metakey (curry #'realm-key *realm*) (car metakeys))) (cdr metakeys)))))
    (do- keys metakeys)))

(finalize-inheritance (find-class 'standard-alpinist))

(defun specified-p (alpinist key-id &aux (key (realm-key *realm* key-id)) (value (the-key-value key (id alpinist))))
  (cond ((null value) nil)
        ((subtypep (key-value-type key) 'string)
         (plusp (length value)))
        ((subtypep (key-value-type key) 'real)
         (not (zerop value)))))

(defun key-value-defaults (key &aux (type (key-value-type key)))
  (cond 
    ((subtypep type 'null) nil)
    ((subtypep type '(or integer form)) 0)
    ((subtypep type 'string) "...")
    (t nil)))

(eval-when (:compile-toplevel :load-toplevel :execute)
  (defun emit-parametrized-alpinist-expression-evaluator-expression (alpinist form declarations)
    `(lambda (,alpinist)
       (declare (special ,alpinist))
       (flet (,@declarations
              (unspecified-p (key-id)
                (not (specified-p ,alpinist key-id))))
         (declare (ignorable ,@(mapcar (lambda (x) `(function ,(first x))) declarations) #'unspecified-p))
         (with-slots (id flag) ,alpinist ,form))))

  (defun make-parametrized-alpinist-expression-evaluator (form alpinist parameters)
    (compile nil (emit-parametrized-alpinist-expression-evaluator-expression alpinist form parameters))))

(defmacro parametrized-alpinist-expression-evaluator (form (alpinist) &rest parameters)
  `(make-parametrized-alpinist-expression-evaluator ,form ',alpinist ',parameters))

(defun make-alpinist-query-predicate (form)
  (parametrized-alpinist-expression-evaluator form (*alpinist*)
    (key (name)
         (realm-key-value *realm* name (id *alpinist*)))))

(defun make-alpinist-query-calculator (form)
  (parametrized-alpinist-expression-evaluator form (*alpinist*)
    (key (name &aux (key (realm-key *realm* name)))
      (if-let ((value (realm-key-value *realm* name (id *alpinist*))))
              (case (key-value-type key)
                (form (funcall (make-alpinist-query-calculator value) *alpinist*))
                (t value))
              (key-value-defaults key)))))

(defun query-realm (query)
  (remove-if-not (make-alpinist-query-predicate query) (map-realm-alpinists #'identity)))

(defmacro with-matches ((a query) &body body)
  (with-gensyms (query-fn)
    `(let ((,query-fn (make-alpinist-query-predicate ,query)))
       (remove nil (map-realm-alpinists (lambda (,a)
                                          (when (funcall ,query-fn ,a)
                                            ,@body)))))))

(defun map-query-realm (query &optional calculative-p)
  (remove-if #'null (map-realm-alpinists (funcall (if calculative-p #'make-alpinist-query-calculator #'make-alpinist-query-predicate) query))))

(defmacro with-calculation ((a query result) &body body)
  (with-gensyms (query-fn)
    `(let ((,query-fn (make-alpinist-query-calculator ,query)))
       (remove nil (map-realm-alpinists (lambda (,a &aux (,result (funcall ,query-fn ,a)))
                                          ,@body))))))

(defun store-realm ()
  (cl-store:store *realm* (realm-storage-pathname *realm*)))

(defun store-realm-as (pathname)
  (cl-store:store *realm* pathname))

(defun restore-realm (pathname)
  (setf *realm* (cl-store:restore pathname)
        (realm-storage-pathname *realm*) pathname))

(defun restore-realm-from-url (url)
  (flexi-streams:with-input-from-sequence (reply (drakma:http-request url))
    (setf *realm* (cl-store:restore reply)
          (realm-storage-pathname *realm*) nil)))

(defvar *recent-realms-store* (merge-pathnames (user-homedir-pathname) (make-pathname :name ".alpine-recent")))
(defvar *recent-realms* nil)

;;;
;;; CLIM
;;;

(setf *tooltip-color* +PERU+
      *tooltip-textstyle* (make-text-style :sans-serif :roman :large))

(defclass conditional-commands-view (view)
  ((desired-commands :accessor view-desired-commands :initarg :desired-commands)
   (undesired-commands :accessor view-undesired-commands :initarg :undesired-commands)))

(defun switch-view (to)
  (setf (stream-default-view *standard-output*) to)
  (dolist (c (view-undesired-commands to))
    (setf (command-enabled c *application-frame*) nil))
  (dolist (c (view-desired-commands to))
    (setf (command-enabled c *application-frame*) t)))

(defclass members-view (conditional-commands-view) 
  ((filter :accessor filter :initform t)
   (selected :accessor selected :initform 0)
   (sort-key :accessor sort-key :initform nil)
   (multiple-columns-p :accessor multiple-columns-p :initform nil)
   (phone-as-tooltips-p :accessor phone-as-tooltips-p :initform t)
   (expressions-as-source-p :accessor expressions-as-source-p :initform nil))
  (:default-initargs
   :desired-commands '(com-list-keyed com-set-filter com-reset-filter)
   :undesired-commands '(com-toggle-keyed)))

(defmethod (setf filter) :around (val (o members-view))
  (block setter
    (handler-case (query-realm val)
      (error () (return-from setter)))
    (call-next-method)
    (let ((string (write-to-string val)))
      (setf (gadget-value (find-pane-named *application-frame* 'filter)) string)
      (format *query-io* "~&Filter set to ")
      (present string 'query :stream *query-io*)
      (finish-output *query-io*)
      val)))

(defun designator-to-query (query)
  "Coerce the query designator to a query."
  (declare (type (or string symbol list) query))
  (typecase query
    ((or symbol list) query)
    (string (ignore-errors (read-from-string query)))))

(defmethod handle-event :around ((o text-editor-pane) (e keyboard-event))
  (format *query-io* "woot~%") (finish-output *query-io*)
;;   (call-next-method)
  t)

(defvar *members-view* (make-instance 'members-view))

(define-application-frame alpine ()
  ()
  (:command-table (alpine-command-table :inherit-from (main-command-table alpinist-command-table key-command-table sort-command-table filter-command-table flag-command-table)))
  (:menu-bar main-command-table)
  (:pointer-documentation t)
  (:panes
   (apply-btn :push-button :name 'apply-button :label "Apply"
              :activate-callback (lambda (btn)
                                   (declare (ignore btn))
                                   (let* ((gadget-value (gadget-value (find-pane-named *application-frame* 'filter)))
                                          (value (if (plusp (length gadget-value)) gadget-value "T")))
                                     (with-output-recording-options (*query-io*)
                                       (execute-frame-command *application-frame* `(com-set-filter ,value)))
                                     (redisplay-frame-pane *application-frame* (find-pane-named *application-frame* 'main-pane) :force-p t))))
   (reset-btn :push-button :name 'reset-button :label "Reset"
              :activate-callback (lambda (btn)
                                   (declare (ignore btn))
                                   (execute-frame-command *application-frame* `(com-reset-filter))
                                   (redisplay-frame-pane *application-frame* (find-pane-named *application-frame* 'main-pane) :force-p t)))
   (display-key-btn :push-button :name 'display-key-button :label "+ Key"
                    :activate-callback (lambda (btn)
                                         (declare (ignore btn))
                                         (when-let* ((choices (metakey- (map-realm-keys #'identity 'standard-key *realm*) (realm-key *realm* 'displayed-keys)))
                                                     (choice (menu-choose (mapcar #'id choices))))
                                           (execute-frame-command *application-frame* `(com-toggle-key-display ,(realm-key *realm* choice))))
                                         (redisplay-frame-pane *application-frame* (find-pane-named *application-frame* 'main-pane) :force-p t)))
   (filter :text-editor :name 'filter :nlines 1 :ncolumns 52 :max-width +fill+ :value (write-to-string (filter *members-view*)))
   (main-pane :application :name 'main-pane :min-height 760 :min-width 1024 :display-function 'display-main-pane :default-view *members-view*)
   (interactor :interactor :min-height 192 :min-width 1024))
  (:layouts
   (default (vertically ()
              (horizontally ()
                display-key-btn reset-btn apply-btn (labelling (:label "Filter:") filter) +fill+)
              main-pane
              interactor))))

(defun notify-error (stream format-string &rest args)
  (surrounding-output-with-border (stream :shape :drop-shadow :background +red+)
    (apply #'format stream format-string args)))

(defun present-print (o s)
  (present o (presentation-type-of o) :stream s))

(defun maybe-invoke-with-surrounding-border (p lambda stream)
  (if p (surrounding-output-with-border (stream :shape :drop-shadow) (funcall lambda)) (funcall lambda)))

(defun present-list (list stream &key description bordered)
  (when description
    (format stream description))
  (maybe-invoke-with-surrounding-border
   bordered (lambda ()
              (filling-output (stream)
                (format-textual-list list #'present-print :stream stream :separator ", " :conjunction " and" :suppress-separator-before-conjunction t)))
   stream))

(defun smarter-write (o s)
  (if (and (stringp o) (zerop (length o)))
      (format s "...")
      (format s "~a" o)))

(defun format-phone (phone &aux (split (and phone (split-number phone '(10 7 4)))))
  (if phone
      (with-output-to-string (s)
        (format s "~D ~3,'0D ~3,'0D ~4,'0D" (first split) (second split) (third split) (fourth split)))
      "..."))

(defun format-passport-id (id &aux (split (split-number id '(6))))
  (format nil "~4,'0D ~6,'0D" (first split) (second split)))

(defun better-name (alpinist)
  (with-output-to-string (s)
    (format s "~A ~A" (alpinist-key-value alpinist 'familyname) (cute-name (alpinist-key-value alpinist 'name)))))

(defmethod print-object ((o standard-alpinist) stream)
  (write (better-name o) :stream stream :escape nil))

(defun render-key-value (stream key-value)
  (let ((key (value-key key-value))
        (value (alpinist-key-value* key-value)))
    (cond ((and (eq (key-value-type key) 'form)
                (expressions-as-source-p *members-view*))
           (let ((alpinist (alpinist (value-key-key key-value))))
             (format stream "~A" (handler-case (funcall (make-alpinist-query-calculator value) alpinist)
                                   (serious-condition (c)
                                     (notify-error *query-io* "bad form ~S for key ~S, alpinist ~A: ~A" value (id key) alpinist c) (terpri *query-io*)
                                     "<bad form>")))))
          (t (smarter-write value stream)))))

(define-presentation-method present (o (type symbol) s view &key)                 (smarter-write o s))
(define-presentation-method present (o (type string) s view &key)                 (smarter-write o s))
(define-presentation-method present (o (type phone-number) s view &key)           (format s (format-phone o)))
(define-presentation-method present (o (type standard-alpinist) s view &key)      (format s "~A" o))
(define-presentation-method present (o (type key) s view &key)                    (format s "~A" o))
(define-presentation-method present (o (type query) s view &key)                  (format s "#<query ~A>" o))
(define-presentation-method present (o (type flag) s view &key)                   (format s "~:[   ~; o ~]" (flag (who o))))
(define-presentation-method present (o (type key-value) s view &key)              (format s "~A: ~A" (description (value-key o)) (alpinist-key-value* o)))
(define-presentation-method present (o (type brief-key-value) s view &key)        (render-key-value s o))
(define-presentation-method present (o (type brief-unset-key-value) s view &key)  (format s "    "))

(define-presentation-method present (o (type passport-id) s view &key)
  (write (format-passport-id o) :stream s))

(define-presentation-method present (o (type date) s view &key)
  (multiple-value-bind (second minute hour day month year day-of-week) (decode-universal-time o)
    (declare (ignore second))
    (format s "~[Mon~;Tue~;Wed~;Thu~;Fri~;Sat~;Sun~] ~[Jan~;Feb~;Mar~;Apr~;May~;Jun~;Jul~;Aug~;Sep~;Oct~;Nov~;Dec~] ~D ~D, ~D:~D" day-of-week month day year hour minute)))

(define-presentation-translator gentler-expression-to-form
    (expression form global-command-table
                :gesture :select
                :menu nil
                :priority 12)
    (object)
  object)

(define-presentation-method highlight-presentation :after ((type standard-alpinist) record stream (state (eql :highlight)))
  (when (phone-as-tooltips-p *members-view*)                          
    (draw-tooltip stream (format-phone (alpinist-key-value (slot-value record 'climi::object) 'phone)) :region (orec-relative->absolute-region record stream))))

(define-presentation-method highlight-presentation :after ((type standard-alpinist) record stream (state (eql :unhighlight)))
  (declare (ignore record))
  (when (phone-as-tooltips-p *members-view*)                          
    (erase-tooltip stream)))

(define-presentation-method highlight-presentation :after ((type key) record stream (state (eql :highlight)))
  (draw-tooltip stream (symbol-name (id (slot-value record 'climi::object))) :region (orec-relative->absolute-region record stream)))

(define-presentation-method highlight-presentation :after ((type key) record stream (state (eql :unhighlight)))
  (declare (ignore record))
  (erase-tooltip stream))

(defgeneric display-pane-with-view (frame pane view))

(defun display-main-pane (frame pane)
  (display-pane-with-view frame pane (stream-default-view pane)))

(defun key-value> (key v1 v2)
  (case (key-value-type key)
    ((nil) (and v1 (not v2)))
    (integer (and v1 (or (not v2) (> v1 v2))))
    (string (and v1 (or (not v2) (string< v1 v2))))))

(defun display-members-view-to-stream (stream view &aux (displayed-keys (realm-key *realm* 'displayed-keys)))
  (formatting-table (stream :multiple-columns (multiple-columns-p view))
    (formatting-row (stream)
      (formatting-cell (stream) (format stream "+/-"))
      (formatting-cell (stream) (format stream "Name"))
      (unless (phone-as-tooltips-p view)
        (formatting-cell (stream) (format stream "Phone")))
      (map-metakey (lambda (id &aux (key (realm-key *realm* id)))
                     (formatting-cell (stream :align-x :center)
                       (with-output-as-presentation (stream key 'standard-key-banner)
                         (write (description key) :stream stream :escape nil))))
                   displayed-keys))
    (iter (with alpinists = (query-realm (filter view)))
          (for a in (if-let ((key (sort-key view)))
                            (sort alpinists (curry #'key-value> key) :key (rcurry #'alpinist-key-value (id key)))
                            (sort alpinists #'string< :key (rcurry #'alpinist-key-value 'familyname))))
          (flet ((present-key-value (key stream)
                   (let ((val (the-key-value key (id a))))
                     (formatting-cell (stream :align-x (if (stringp val) :left :center))
                       (present (make-instance (if val 'brief-key-value 'brief-unset-key-value) :key key :key-key (id a))
                                (if val 'brief-key-value 'brief-unset-key-value)
                                :stream stream)))))
            (formatting-row (stream)
              (formatting-cell (stream :align-x :center)
                (present (make-instance 'flag :who a) 'flag :stream stream))
              (with-output-as-presentation (stream a 'standard-alpinist :background +red+)
                (formatting-cell (stream)
                  (write (better-name a) :stream stream :escape nil))
                (unless (phone-as-tooltips-p view)
                  (present-key-value (realm-key *realm* 'phone) stream))
                (map-metakey (lambda (id)
                               (present-key-value (realm-key *realm* id) stream))
                             displayed-keys))))))
  (change-space-requirements stream
                             :width (bounding-rectangle-width (stream-output-history stream))
                             :height (bounding-rectangle-height (stream-output-history stream))))

(defmethod display-pane-with-view (frame pane (view members-view))
  (display-members-view-to-stream pane view))

(defclass alpinist-view (conditional-commands-view)
  ((%alpinist :initarg :alpinist :reader view-alpinist))
  (:default-initargs
   :desired-commands '(com-toggle-keyed)
   :undesired-commands '(com-list-keyed com-set-filter com-reset-filter)))

(defmethod display-pane-with-view (frame pane (view alpinist-view))
  (let ((alpinist (view-alpinist view))
        (personal (realm-key *realm* 'personal)))
    (flet ((present-personal-key (key)
             (formatting-row (pane)
               (formatting-cell (pane) (write (description key) :stream pane :escape nil))
               (formatting-cell (pane) (present (make-instance 'brief-key-value :key-key (id alpinist) :key key) 'brief-key-value :stream pane :sensitive t)))))
      (formatting-table (pane)
        (map-metakey (compose #'present-personal-key (curry #'realm-key *realm*)) personal)))
    (present-list (mapcar (curry #'make-instance 'key-value :key-key (id alpinist) :key) (metakey- (alpinist-member-keys (id alpinist)) personal))
                  pane :description "~&Member keys~16,1T")
    (present-list (mapcar (curry #'make-instance 'unset-key-value :key-key (id alpinist) :key) (metakey- (alpinist-non-member-keys (id alpinist)) personal))
                  pane :description "~&Non-member keys~16,1T")))

;;;
;;; Commands: main-command-table
;;;

(define-command-table realm-command-table
    :menu (("Open realm" :command com-open-realm)
           ("Open realm from URL" :command com-open-realm-url)
           ("Open recent realm" :command com-open-recent-realm)
           ("Store realm" :command com-store-realm)
           ("Store realm as" :command com-store-realm-as)))

(defun add-recent-realm-storage-location (pathname)
  (setf (command-enabled 'com-open-recent-realm *application-frame*) t)
  (pushnew pathname *recent-realms* :test #'equal))

(define-alpine-command (com-open-realm :name t :command-table realm-command-table) ((path 'pathname))
  (add-recent-realm-storage-location path)
  (restore-realm path))

(define-alpine-command (com-open-realm-url :name t :command-table realm-command-table) ((url 'url))
  (add-recent-realm-storage-location url)
  (restore-realm-from-url url))

(define-alpine-command (com-open-recent-realm :name t :command-table realm-command-table) ()
  (when-let ((spec (menu-choose *recent-realms*)))
    (if (pathnamep spec) (restore-realm spec) (restore-realm-from-url spec))))

(define-alpine-command (com-store-realm :name t :command-table realm-command-table) ()
  (if (realm-storage-pathname *realm*)
      (store-realm)
      (notify-error *query-io* "Cannot write realm back to an URL, use the 'Save As' command.")))

(define-alpine-command (com-store-realm-as :name t :command-table realm-command-table) ()
  (present-list (remove-if-not #'pathnamep *recent-realms*) *query-io* :description "Recent realms:~%" :bordered t)
  (terpri *query-io*)
  (let ((path (accept 'pathname)))
    (add-recent-realm-storage-location path)
    (store-realm-as path)))

(define-command-table filter-command-table
    :menu (("Set filter" :command com-set-filter)
           ("Narrow filter" :command com-narrow-filter)
           ("Set filter to key" :command com-set-filter-to-key)
           ("Reset filter" :command com-reset-filter)))

(defun view-filter ()
  (filter (stream-default-view *standard-output*)))

(defun (setf view-filter) (query)
  (setf (filter (stream-default-view *standard-output*)) query))

(define-alpine-command (com-set-filter :name t :command-table filter-command-table) ((query 'query))
  (unless (when-let ((query (designator-to-query query)))
            (setf (view-filter) query))
    (notify-error *query-io* "Bad query '~A'~%" query)))

(define-alpine-command (com-narrow-filter :name t :command-table filter-command-table) ((query 'query))
  (unless (when-let ((query (designator-to-query query)))
            (setf (view-filter) (list 'and (filter (stream-default-view *standard-output*)) query)))
    (notify-error *query-io* "Bad subquery '~A'~%" query)))

(define-alpine-command (com-set-filter-to-key :name t :command-table filter-command-table) ((key 'standard-key))
  (setf (view-filter) `(key ',(id key))))

(define-alpine-command (com-reset-filter :name t :command-table filter-command-table) ()
  (setf (view-filter) t))

(define-command-table sort-command-table
    :menu (("Sort by key" :command com-sort-by-key)
           ("Sort by name" :command com-sort-by-name)))

(define-alpine-command (com-sort-by-key :name t :command-table sort-command-table) ((key 'standard-key :prompt "Sort by" :gesture :menu))
  (setf (sort-key *members-view*) key))

(define-alpine-command (com-sort-by-name :name t :command-table sort-command-table) ()
  (setf (sort-key *members-view*) nil))

(define-command-table key-command-table
    :menu (("Add key" :command com-add-key)
;;            ("Add metakey" :command com-add-metakey)
           ("Remove key" :command com-remove-key)
           ("Toggle key display" :command com-key-toggle-display)
           ("Toggle key value typedness" :command com-key-toggle-value-typedness)
           ("Add key to flags" :command com-add-key-to-flags)
           ("Set key value for flagged" :command com-set-key-value-flagged)
           ("List keys" :command com-list-keys)
           ("List metakeys" :command com-list-metakeys)
           ("List keyed" :command com-list-keyed)
           ("Query" :command com-query-realm)
           ("Set displayed keys" :command com-set-displayed-keys)))

(define-alpine-command (com-add-key :name t :command-table key-command-table) ()
  (realm-add-key *realm* 'standard-key (accept 'symbol :stream *query-io* :prompt "Key ID") (accept 'string :stream *query-io* :prompt "Description")))

;; (define-alpine-command (com-add-metakey :name t :command-table key-command-table) ()
;;   (let ((metakey (realm-add-key *realm* 'metakey (accept 'symbol :stream *query-io* :prompt "Key ID") (accept 'string :stream *query-io* :prompt "Description")))
;;         (key-ids (mapcar #'id (accept '(sequence standard-key)))))
;;     (setf (the-key-values metakey key-ids) (make-list (length key-ids) :initial-element t))))

(define-alpine-command (com-remove-key :name t :command-table key-command-table) ((key 'standard-key :prompt "Key"))
  (when (yes-or-no-p "Really remove key ~S?" key)
    (remove-realm-object *realm* (id key))))

(define-alpine-command (com-toggle-key-display :name t :command-table key-command-table) ((key 'standard-key :prompt "Key"))
  (notf (the-key-value (realm-key *realm* 'displayed-keys) (id key))))

(define-alpine-command (com-add-key-to-flags :name t :command-table key-command-table) ((key 'standard-key :prompt "Key" :gesture :menu))
  (map-realm-alpinists (lambda (a)
                         (orf (flag a) (the-key-value key (id a))))))

(define-command (com-set-key-value-flagged :name t :command-table key-command-table) ((key 'standard-key-banner :gesture :menu))
  (let ((value (if-let ((type (key-value-type key)))
                       (get-key-value type)
                       t)))
    (with-matches (a (filter *members-view*))
      (when (flag a)
        (setf (the-key-value key (id a)) value)))))

(define-alpine-command (com-key-toggle-value-typedness :name t :command-table key-command-table) ((key 'standard-key :prompt "Key"))
  (setf (key-value-type key) (if (key-value-type key) nil (menu-choose '(nil string integer form)))))

(define-alpine-command (com-list-keys :name t :command-table key-command-table) ()
  (present-list (map-realm-keys #'identity 'standard-key *realm*) *query-io* :description "~&Known keys:~%" :bordered t))

(define-alpine-command (com-list-metakeys :name t :command-table key-command-table) ()
  (present-list (map-realm-keys #'identity 'metakey *realm*) *query-io* :description "~&Known metakeys:~%" :bordered t))

(defun sort-people (people)
  (sort people #'string< :key (rcurry #'alpinist-key-value 'familyname)))

(define-alpine-command (com-list-keyed :name t :command-table key-command-table) ((key 'standard-key :gesture :menu))
  (format *query-io* "~&People matching ") (present key 'standard-key :stream *query-io*)
  (present-list (sort-people (remove-if-not (rcurry #'alpinist-key-value (id key)) (map-realm-alpinists #'identity))) *query-io*
                :description ":~%" :bordered t))

(define-alpine-command (com-query-realm :name t :command-table key-command-table) ((query 'query))
  (when-let ((query (designator-to-query query)))
    (format *query-io* "~&People matching ") (present query 'query :stream *query-io*)
    (present-list (sort-people (query-realm query)) *query-io* :description ":~%" :bordered t)))

(defun select-subset (items stream &aux check-box)
  (accepting-values (stream :own-window t)
    (with-output-as-gadget (stream :parent stream)
      (setf check-box (make-pane 'check-box :client stream :id 'check-box))
      (iter (for (id desc value) in items)
            (make-pane 'toggle-button :id id :label desc :value value :parent check-box))))
  (mapcar #'gadget-id (remove-if-not #'gadget-value (sheet-children check-box))))

(define-alpine-command (com-set-displayed-keys :name t :command-table key-command-table) ()
;;   (setf (displayed-keys *members-view*)
;;         (select-subset (map-realm-keys (lambda (k) (list k (description k) (not (null (member k (displayed-keys *members-view*)))))) *realm*)
;;                        *query-io*))
    (let ((select-hash (make-hash-table))
          (displayed-keys (realm-key *realm* 'displayed-keys)))
      (accepting-values (*query-io* :own-window t)
        (map-realm-keys (lambda (k)
                          (setf (gethash k select-hash)
                                (accept 'symbol :prompt (description k) :stream *query-io* :default (the-key-value displayed-keys (id k)) :insert-default t))
                          (terpri *query-io*))
                        'standard-key
                        *realm*))
      (iter (for (k v) in-hashtable select-hash)
            (setf (the-key-value displayed-keys (id k)) (not (null v)))))
  )

(define-presentation-to-command-translator set-filter
    (query com-set-filter key-command-table :documentation "Query" :pointer-documentation "Query")
    (object)
  (list object))

(define-presentation-to-command-translator set-filter-to-key
    (standard-key-banner com-set-filter-to-key key-command-table :documentation "Set filter to" :pointer-documentation "Set filter to" :echo nil)
    (object)
  (list object))

(define-presentation-to-command-translator toggle-key-display
    (standard-key com-toggle-key-display key-command-table :documentation "Toggle display" :pointer-documentation "Toggle display" :echo nil)
    (object)
  (list object))

(define-command-table flag-command-table
    :menu (("Clear flags" :command com-clear-flags)
           ("Invert flags" :command com-invert-flags)
           ("Toggle flag" :command com-toggle-flag)))

(define-alpine-command (com-clear-flags :name t :command-table flag-command-table) ()
  (map-realm-alpinists (lambda (a)
                         (setf (flag a) nil))))

(define-alpine-command (com-invert-flags :name t :command-table flag-command-table) ()
  (map-realm-alpinists (lambda (a)
                         (notf (flag a)))))

(define-alpine-command (com-toggle-flag :name t :command-table flag-command-table) ((flag 'flag))
  (setf (flag (who flag)) (not (flag (who flag)))))

(define-presentation-to-command-translator toggle-flag
    (flag com-toggle-flag flag-command-table :documentation "Toggle" :pointer-documentation "Toggle" :echo nil)
    (object)
  (list object))

(define-command-table config-command-table
    :menu (("Phone as tooltips" :command com-toggle-phone-as-tooltips)
           ("Members in multiple columns" :command com-toggle-multiple-columns)
           ("Expressions as source" :command com-toggle-expressions-as-source)))

(define-alpine-command (com-toggle-phone-as-tooltips :name t :command-table config-command-table) ()
  (notf (slot-value *members-view* 'phone-as-tooltips-p)))

(define-alpine-command (com-toggle-multiple-columns :name t :command-table config-command-table) ()
  (notf (slot-value *members-view* 'multiple-columns-p)))

(define-alpine-command (com-toggle-expressions-as-source :name t :command-table config-command-table) ()
  (notf (slot-value *members-view* 'expressions-as-source-p)))

(define-command-table main-command-table
    :menu (("Realm" :menu realm-command-table)
           ("Show all" :command com-show-all)
           ("Write PS" :command com-write-ps)
           ("Define new" :command com-define-alpinist)
           ("Flags" :menu flag-command-table)
           ("Keys" :menu key-command-table)
           ("Sort" :menu sort-command-table)
           ("Filter" :menu filter-command-table)
           ("Config" :menu config-command-table)
           ("Quit" :command com-quit)))

(define-command (com-show-all :name t :command-table main-command-table) ()
  (switch-view *members-view*))

(define-command (com-define-alpinist :name t :command-table main-command-table) ()
  (let ((id (accept 'symbol :prompt "Id")))
    (unless (alpinist id :if-does-not-exist :continue)
      ;;       (format pane "Alpinist by name ~S is already defined.~%" name)
      (let ((familyname (accept 'string :prompt "Family name"))
            (name (accept 'string :prompt "Name"))
            (surname (accept 'string :prompt "Surname"))
            (alpinist (make-instance 'standard-alpinist :id id)))
        (setf (alpinist id) alpinist
              (alpinist-key-value alpinist 'name) name
              (alpinist-key-value alpinist 'surname) surname
              (alpinist-key-value alpinist 'familyname) familyname
              (alpinist-key-value alpinist 'phone) (accept 'integer :prompt "Phone number"))))))

(define-command (com-show-alpinist :name t :command-table main-command-table) ((alpinist 'standard-alpinist))
  (switch-view (make-instance 'alpinist-view :alpinist alpinist)))

(define-command (com-write-ps :name t :command-table main-command-table) ((path 'pathname))
  (with-open-file (file-stream path :direction :output :if-does-not-exist :create :if-exists :supersede)
    (with-output-to-postscript-stream (stream file-stream)
      (display-members-view-to-stream stream *members-view*))))

(define-command (com-quit :name t :command-table main-command-table) ()
  (frame-exit *application-frame*))

(define-alpine-command (com-previous-alpinist :name t :command-table main-command-table :keystroke :up) ()
  (let ((total (hash-table-count (realm-alpinists *realm*))))
    (when (plusp total)
      (setf (selected (stream-default-view *standard-output*)) (mod (1- (selected (stream-default-view *standard-output*))) total)))))

(define-alpine-command (com-next-alpinist :name t :command-table main-command-table :keystroke :down) ()
  (let ((total (hash-table-count (realm-alpinists *realm*))))
    (when (plusp total)
      (setf (selected (stream-default-view *standard-output*)) (mod (1+ (selected (stream-default-view *standard-output*))) total)))))

(define-presentation-to-command-translator show-alpinist
    (standard-alpinist com-show-alpinist main-command-table :documentation "Show alpinist" :pointer-documentation "Show alpinist")
    (object)
  (list object))

;;;
;;; Commands: alpinist-command-table
;;;
(define-command-table alpinist-command-table
    :menu ())

(defun get-key-value (type &optional default)
  (if type (accept type :default default :insert-default default) t))

(define-command (com-set-key-value :name t :command-table alpinist-command-table) ((key-value 'unset-key-value))
  (setf (alpinist-key-value* key-value) (get-key-value (key-value-type (value-key key-value)) (or (alpinist-key-value* key-value) (key-value-defaults (value-key key-value)))))
  (change-class key-value 'key-value))

(defun key-value-unset (key-value)
  (change-class key-value 'unset-key-value)
  (notf (alpinist-key-value* key-value)))

(define-command (com-edit-key-value :name t :command-table alpinist-command-table) ((key-value 'key-value))
  (if-let ((type (key-value-type (value-key key-value))))
          (setf (alpinist-key-value* key-value) (get-key-value type (or (alpinist-key-value* key-value) (key-value-defaults (value-key key-value)))))
          (key-value-unset key-value)))

(define-command (com-unset-key-value :name t :command-table alpinist-command-table) ((key-value 'key-value :gesture :menu))
  (key-value-unset key-value))

(define-presentation-to-command-translator set-key-value
    (unset-key-value com-set-key-value alpinist-command-table :documentation "Toggle" :pointer-documentation "Toggle")
    (object)
  (list object))

(define-presentation-to-command-translator edit-key-value
    (key-value com-edit-key-value alpinist-command-table :documentation "Edit" :pointer-documentation "Edit")
    (object)
  (list object))

(defun ui ()
  (let ((frame (make-application-frame 'alpine)))
    (unless (setf *recent-realms* (with-open-file (s *recent-realms-store* :if-does-not-exist :create)
                                    (remove-if-not (lambda (x) (or (probe-file x) (stringp x))) (read s nil nil))))
      (setf (command-enabled 'com-open-recent-realm frame) nil))
    (run-frame-top-level frame))
  (with-open-file (s *recent-realms-store* :direction :output :if-exists :supersede :if-does-not-exist :create)
    (print *recent-realms* s)))
